#include <omp.h>
#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <inttypes.h>

#define NUM_RUNS 1
#define NUM_INCS 10000
#define MAX_THREADS 32

struct __attribute__((aligned(64))) __attribute__ ((__packed__)) element {
	uint32_t value;
	char padding[60];
};

void print_stats(char* lockname, int numthreads, double usecs) {
    static char first = 1;
    if (first) {
        printf("%10s %8s %8s\n", "lockname", "threads", "usecs");
        first = 0;
    }
    printf("%10s %8i %8.4lf\n", lockname, numthreads, usecs);
}

double diff(struct timespec start, struct timespec end) {

    struct timespec temp;
    double res;

    if ((end.tv_nsec-start.tv_nsec)<0) {
        temp.tv_sec = end.tv_sec-start.tv_sec-1;
        temp.tv_nsec = 1e9+end.tv_nsec-start.tv_nsec;
    } else {
        temp.tv_sec = end.tv_sec-start.tv_sec;
        temp.tv_nsec = end.tv_nsec-start.tv_nsec;
    }
    res = ((double) temp.tv_sec) * 1e6 + ((double) temp.tv_nsec) / 1e3;

    return res;

}

int benchmark_array_lock_opt() {

   struct timespec time1, time2;
   volatile int* a = malloc(sizeof(int));

    for (int threads=1; threads<=MAX_THREADS; threads++) {

        volatile struct element* flag = malloc(sizeof(struct element) * threads);
        volatile int* tail = malloc(sizeof(int));

        for (int run=0; run<NUM_RUNS; run++) {

            *a = 0;
            for (int i=0; i<threads; i++) flag[i].value = 0;
            flag[0].value = 1;
            *tail = 0;

            clock_gettime(CLOCK_REALTIME, &time1);
            #pragma omp parallel shared(flag,tail,a) num_threads(threads)
            {
              int me = omp_get_thread_num();
	          int index = 0;
              
                for (int i=0; i<NUM_INCS; i++) {
                    // lock
                    index = __sync_fetch_and_add(tail, 1) % threads;
                    while (!flag[index].value) {};

                    // CR
                    *a += 1;

                    // unlock
                    __asm__("mfence");
                    flag[index].value = 0;
                    flag[(index + 1) % threads].value = 1;
                }
            }

            clock_gettime(CLOCK_REALTIME, &time2);
            print_stats("ARR_OPT", threads, diff(time1, time2)/NUM_INCS);

            if(*a != threads*NUM_INCS) {
                fprintf(stderr, "Wrong result! [%i]\n", *a);
                exit(EXIT_FAILURE);
            }
        }

        free((void*) flag);
        free((void*) tail);

    }

    free((void*)a);

}

int main(int argc, char** argv) {

    benchmark_array_lock_opt();

}



